package meerkat;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.ProtobufsMessages.BasicMessage;
import meerkat.ProtobufsMessages.Tag;
import meerkat.bulletinboard.SimpleBulletinBoardClient;
import meerkat.comm.CommunicationException;
import meerkat.crypto.Encryption;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.Crypto;
import util.AccurateTimestamp;
import util.RegistryTagTypes;

import java.io.IOException;
import java.util.*;

/**
 *  Created by Vladimir Eliezer Tokarev on 1/8/2016.
 *  Gives the ability to manage voters information
 *  This object is synchronous (which meaning its method blocking)
 */
public class SimpleRegistry {

    protected Encryption signatory;

    protected SimpleBulletinBoardClient communicator;

    /**
     * @param signatory implements the DigitalSignature interface
     * @param communicator implements the BulletinBoardClient interface
     */
    public SimpleRegistry(Encryption signatory, SimpleBulletinBoardClient communicator){
        this.signatory = signatory;
        this.communicator = communicator;
    }

    /**
     * Adds new voter to the bulletin-board
     * @param voterID
     * @param personalData for example residence location
     * @throws throws CommunicationException
     * @return void
     */
    public void AddVoter(String voterID, String personalData) throws CommunicationException, IOException {
        Tag.Builder idTag = Tag.newBuilder();
        idTag.setContent(RegistryTagTypes.ID_TAG + " " + voterID);

        Tag.Builder voterEntryTag = Tag.newBuilder();
        voterEntryTag.setContent(RegistryTagTypes.VOTER_ENTRY_TAG);

        Tag.Builder timestampTag = Tag.newBuilder();
        timestampTag.setContent(RegistryTagTypes.ACTION_TIMESTAMP_TAG + " "
                + AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage =
                BasicMessage.newBuilder().addTag(idTag).addTag(voterEntryTag).addTag(timestampTag);
        basicMessage.setData(ByteString.copyFrom(personalData.getBytes()));

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Adding given voter to given group
     * @param voterID
     * @param groupID
     * @throws CommunicationException
     * @return true if the adding action succeeded else return false
     */
    public void AddToGroup(String voterID, String groupID) throws CommunicationException, IOException {
        Tag.Builder idTag = Tag.newBuilder();
        idTag.setContent(RegistryTagTypes.ID_TAG + " " + voterID);

        Tag.Builder groupIDTag = Tag.newBuilder();
        groupIDTag.setContent(RegistryTagTypes.GROUP_ID_TAG + " " + groupID);

        Tag.Builder actionTag = Tag.newBuilder();
        actionTag.setContent(RegistryTagTypes.GROUP_ACTION_TAG + " " + RegistryTagTypes.ADD_TO_GROUP_TAG);

        Tag.Builder timestampTag = Tag.newBuilder();
        timestampTag.setContent(RegistryTagTypes.ACTION_TIMESTAMP_TAG + " "
                +AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage =
                BasicMessage.newBuilder().addTag(idTag).addTag(groupIDTag).addTag(actionTag).addTag(timestampTag);

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Removes given voter from given group
     * @param voterID
     * @param groupID
     * @return  true if the removing action succeeded else return false
     * @throws CommunicationException
     */
    public void RemoveFromGroup(String voterID, String groupID) throws CommunicationException, IOException {
        Tag.Builder idTag = Tag.newBuilder();
        idTag.setContent(RegistryTagTypes.ID_TAG + " " + voterID);

        Tag.Builder groupIDTag = Tag.newBuilder();
        groupIDTag.setContent(RegistryTagTypes.GROUP_ID_TAG + " " + groupID);

        Tag.Builder actionTag = Tag.newBuilder();
        actionTag.setContent(RegistryTagTypes.GROUP_ACTION_TAG + " " + RegistryTagTypes.REMOVE_FROM_GROUP_TAG);

        Tag.Builder timestampTag = Tag.newBuilder();
        timestampTag.setContent(RegistryTagTypes.ACTION_TIMESTAMP_TAG + " "
                +AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage =
                BasicMessage.newBuilder().addTag(idTag).addTag(groupIDTag).addTag(actionTag).addTag(timestampTag);

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Sets that the voter have voted
     * @param id
     * @return true if the set voted succeded else false
     * @throws CommunicationException
     */
    public void SetVoted(String id) throws CommunicationException, IOException {
        Tag.Builder idTag = Tag.newBuilder();
        idTag.setContent(RegistryTagTypes.ID_TAG + " " + id);

        Tag.Builder voteAction = Tag.newBuilder();
        voteAction.setContent(RegistryTagTypes.VOTE_ACTION_TAG);

        Tag.Builder timestampTag = Tag.newBuilder();
        timestampTag.setContent(RegistryTagTypes.ACTION_TIMESTAMP_TAG + " "
                +AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage = BasicMessage.newBuilder().addTag(idTag).addTag(voteAction).addTag(timestampTag);

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Creates bulletin board message from basic message (with the signatures parts)
     * @param basicMessage  BasicMessage
     * @return BulletinBoardAPI.BulletinBoardMessage
     */
    private BulletinBoardAPI.BulletinBoardMessage CreateBulletinBoardMessage(BasicMessage  basicMessage) throws IOException {

        BulletinBoardAPI.BulletinBoardMessage.Builder bulletinBoardMessage =
                BulletinBoardAPI.BulletinBoardMessage.newBuilder();

        // signs the basic message
        Crypto.RerandomizableEncryptedMessage encryptedMessage =
                signatory.encrypt(basicMessage, signatory.generateRandomness(new Random()));

        // creates the signature of signed basic message
        Crypto.Signature.Builder messageSignature = Crypto.Signature.newBuilder();
        messageSignature.mergeFrom(encryptedMessage);

        // creates the unsigned basic message
        BulletinBoardAPI.UnsignedBulletinBoardMessage.Builder unsignedBulletinBoardMessage =
                BulletinBoardAPI.UnsignedBulletinBoardMessage.newBuilder();
        unsignedBulletinBoardMessage.setData(basicMessage.toByteString());

        // sets the signature and the byte array
        bulletinBoardMessage.addSig(messageSignature);

        bulletinBoardMessage.setMsg(unsignedBulletinBoardMessage);

        return bulletinBoardMessage.build();
    }

    /**
     * Gets messages that have the wanted id tag and have or the ADD_TO_GROUP_TAG or REMOVE_FROM_GROUP_TAG
     * @param id
     * @return  List<BulletinBoardAPI.BulletinBoardMessage>
     */
    private List<BulletinBoardAPI.BulletinBoardMessage> GetRelevantMessages(String id) {
        BulletinBoardAPI.MessageFilterList.Builder filters =
                BulletinBoardAPI.MessageFilterList.newBuilder();

        BulletinBoardAPI.MessageFilter.Builder idFilter =
                BulletinBoardAPI.MessageFilter.newBuilder().setTag(RegistryTagTypes.ID_TAG + " " + id)
                        .setTag(RegistryTagTypes.GROUP_ACTION_TAG + " " + RegistryTagTypes.ADD_TO_GROUP_TAG)
                        .setTag(RegistryTagTypes.GROUP_ACTION_TAG + " " + RegistryTagTypes.REMOVE_FROM_GROUP_TAG);

        filters.addFilter(idFilter);

        return communicator.readMessages(filters.build());
    }

    /**
     * Gets all the basic messages from bulletin board messages
     * @param listOfMessages
     * @return List<BasicMessage> G
     * @throws InvalidProtocolBufferException
     */
    private List<BasicMessage> GetBasicMessagesFromBulletinBoardMessages(
            List<BulletinBoardAPI.BulletinBoardMessage> listOfMessages) throws InvalidProtocolBufferException {

        List<BasicMessage> basicMessages = new ArrayList<BasicMessage>();

        for (BulletinBoardAPI.BulletinBoardMessage bulletinBoardMessage : listOfMessages){
           BasicMessage.Builder basicMessage =
                   BasicMessage.newBuilder().mergeFrom(bulletinBoardMessage.getMsg().getData());
           basicMessages.add(basicMessage.build());
        }

        return basicMessages;
    }


    private List<String> GetAllGroupsIdsUserIn(List<BasicMessage> messages){
        Map<String, BasicMessage> groupIdToMessage = new HashMap<String, BasicMessage>();
        List<String> latestStateOfGroup = new ArrayList<String>();

        for (BasicMessage message : messages) {
            BasicMessage temporary = groupIdToMessage.get(message.get)
        }
    }


    /**
     * Requests all the groups that the given id voter is in
     * @param id
     * @return list of groups ids (or names), if the method fails its empty
     * @throws CommunicationException
     */
    public List<String> GetGroups(String id)  throws CommunicationException {
        List<BulletinBoardAPI.BulletinBoardMessage> relevantMessages = GetRelevantMessages(id);
        List<String> groups = new ArrayList<String>();



        /**
         *  Retrieve all List<BulletinBoardMessage> that contains this id
         *  creates new list of strings with the names of the group
         *  Returns the created list
         */
    }

    /**
     * Retreives list of strings that represents voter
     * @param id
     * @return list of strings (empty list if the lookup failed)
     * @throws CommunicationException
     */
    public List<String> GetPersonIDDetails(String id) throws CommunicationException {
        /**
         *      Retreive all List<BulletinBoardMessage> that contains this id
         *      search for message with  the wanted personal data
         *      Create list with personal data
         *      return the list
         */
    }
}
